@@ -23,83 +23,56 @@ import sys as _coconut_sys
 # VERSION:
 # -----------------------------------------------------------------------------------------------------------------------
 
-VERSION = "2.1.1"
-VERSION_NAME = "The Spanish Inquisition"
+VERSION = "3.0.4"
+VERSION_NAME = None
 # False for release, int >= 1 for develop
-DEVELOP = 21
+DEVELOP = 8
 ALPHA = False  # for pre releases rather than post releases
 
+assert DEVELOP is False or DEVELOP >= 1, "DEVELOP must be False or an int >= 1"
+assert DEVELOP or not ALPHA, "alpha releases are only for develop"
+
 # -----------------------------------------------------------------------------------------------------------------------
 # UTILITIES:
 # -----------------------------------------------------------------------------------------------------------------------
 
 
-def _indent(code, by=1, tabsize=4, newline=False, strip=False):
+def _indent(code, by=1, tabsize=4, strip=False, newline=False, initial_newline=False):
     """Indents every nonempty line of the given code."""
-    return "".join(
+    return ("\n" if initial_newline else "") + "".join(
         (" " * (tabsize * by) if line.strip() else "") + line
         for line in (code.strip() if strip else code).splitlines(True)
     ) + ("\n" if newline else "")
 
 
+def _get_target_info(target):
+    """Return target information as a version tuple."""
+    if not target or target == "universal":
+        return ()
+    elif len(target) == 1:
+        return (int(target),)
+    else:
+        return (int(target[0]), int(target[1:]))
+
+
 # -----------------------------------------------------------------------------------------------------------------------
-# CONSTANTS:
+# HEADER:
 # -----------------------------------------------------------------------------------------------------------------------
 
-assert isinstance(DEVELOP, int) or DEVELOP is False, "DEVELOP must be an int or False"
-assert DEVELOP or not ALPHA, "alpha releases are only for develop"
-
-if DEVELOP:
-    VERSION += "-" + ("a" if ALPHA else "post") + "_dev" + str(int(DEVELOP))
-VERSION_STR = VERSION + " [" + VERSION_NAME + "]"
-
-PY2 = _coconut_sys.version_info < (3,)
-PY26 = _coconut_sys.version_info < (2, 7)
-PY37 = _coconut_sys.version_info >= (3, 7)
-
-_non_py37_extras = r'''def _coconut_default_breakpointhook(*args, **kwargs):
-    hookname = _coconut.os.getenv("PYTHONBREAKPOINT")
-    if hookname != "0":
-        if not hookname:
-            hookname = "pdb.set_trace"
-        modname, dot, funcname = hookname.rpartition(".")
-        if not dot:
-            modname = "builtins" if _coconut_sys.version_info >= (3,) else "__builtin__"
-        if _coconut_sys.version_info >= (2, 7):
-            import importlib
-            module = importlib.import_module(modname)
-        else:
-            import imp
-            module = imp.load_module(modname, *imp.find_module(modname))
-        hook = _coconut.getattr(module, funcname)
-        return hook(*args, **kwargs)
-if not hasattr(_coconut_sys, "__breakpointhook__"):
-    _coconut_sys.__breakpointhook__ = _coconut_default_breakpointhook
-def breakpoint(*args, **kwargs):
-    return _coconut.getattr(_coconut_sys, "breakpointhook", _coconut_default_breakpointhook)(*args, **kwargs)
-'''
-
 # if a new assignment is added below, a new builtins import should be added alongside it
-_base_py3_header = r'''from builtins import chr, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, repr
-py_chr, py_hex, py_input, py_int, py_map, py_object, py_oct, py_open, py_print, py_range, py_str, py_super, py_zip, py_filter, py_reversed, py_enumerate, py_repr = chr, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, repr
-_coconut_py_str, _coconut_py_super = str, super
+_base_py3_header = r'''from builtins import chr, dict, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, repr
+py_chr, py_dict, py_hex, py_input, py_int, py_map, py_object, py_oct, py_open, py_print, py_range, py_str, py_super, py_zip, py_filter, py_reversed, py_enumerate, py_repr = chr, dict, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, repr
+_coconut_py_str, _coconut_py_super, _coconut_py_dict = str, super, dict
 from functools import wraps as _coconut_wraps
 exec("_coconut_exec = exec")
 '''
 
-PY37_HEADER = _base_py3_header + r'''py_breakpoint = breakpoint
-'''
-
-PY3_HEADER = _base_py3_header + r'''if _coconut_sys.version_info < (3, 7):
-''' + _indent(_non_py37_extras) + r'''else:
-    py_breakpoint = breakpoint
-'''
-
 # if a new assignment is added below, a new builtins import should be added alongside it
-PY27_HEADER = r'''from __builtin__ import chr, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, raw_input, xrange, repr, long
-py_chr, py_hex, py_input, py_int, py_map, py_object, py_oct, py_open, py_print, py_range, py_str, py_super, py_zip, py_filter, py_reversed, py_enumerate, py_raw_input, py_xrange, py_repr = chr, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, raw_input, xrange, repr
-_coconut_py_raw_input, _coconut_py_xrange, _coconut_py_int, _coconut_py_long, _coconut_py_print, _coconut_py_str, _coconut_py_super, _coconut_py_unicode, _coconut_py_repr = raw_input, xrange, int, long, print, str, super, unicode, repr
+_base_py2_header = r'''from __builtin__ import chr, dict, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, raw_input, xrange, repr, long
+py_chr, py_dict, py_hex, py_input, py_int, py_map, py_object, py_oct, py_open, py_print, py_range, py_str, py_super, py_zip, py_filter, py_reversed, py_enumerate, py_raw_input, py_xrange, py_repr = chr, dict, hex, input, int, map, object, oct, open, print, range, str, super, zip, filter, reversed, enumerate, raw_input, xrange, repr
+_coconut_py_raw_input, _coconut_py_xrange, _coconut_py_int, _coconut_py_long, _coconut_py_print, _coconut_py_str, _coconut_py_super, _coconut_py_unicode, _coconut_py_repr, _coconut_py_dict = raw_input, xrange, int, long, print, str, super, unicode, repr, dict
 from functools import wraps as _coconut_wraps
+from collections import Sequence as _coconut_Sequence
 from future_builtins import *
 chr, str = unichr, unicode
 from io import open
@@ -109,15 +82,11 @@ class object(object):
         eq = self == other
         return _coconut.NotImplemented if eq is _coconut.NotImplemented else not eq
     def __nonzero__(self):
-        self_bool = _coconut.getattr(self, "__bool__", None)
-        if self_bool is not None:
-            try:
-                result = self_bool()
-            except _coconut.NotImplementedError:
-                pass
-            else:
-                if result is not _coconut.NotImplemented:
-                    return result
+        if _coconut.hasattr(self, "__bool__"):
+            got = self.__bool__()
+            if not _coconut.isinstance(got, _coconut.bool):
+                raise _coconut.TypeError("__bool__ should return bool, returned " + _coconut.type(got).__name__)
+            return got
         return True
 class int(_coconut_py_int):
     __slots__ = ()
@@ -138,6 +107,8 @@ class range(object):
         return _coconut.reversed(self._xrange)
     def __len__(self):
         return _coconut.len(self._xrange)
+    def __bool__(self):
+        return _coconut.bool(self._xrange)
     def __contains__(self, elem):
         return elem in self._xrange
     def __getitem__(self, index):
@@ -191,7 +162,6 @@ class range(object):
         return self.__class__(*self._args)
     def __eq__(self, other):
         return self.__class__ is other.__class__ and self._args == other._args
-from collections import Sequence as _coconut_Sequence
 _coconut_Sequence.register(range)
 @_coconut_wraps(_coconut_py_print)
 def print(*args, **kwargs):
@@ -238,9 +208,111 @@ def _coconut_exec(obj, globals=None, locals=None):
     if globals is None:
         globals = _coconut_sys._getframe(1).f_globals
     exec(obj, globals, locals)
-''' + _non_py37_extras
+import operator as _coconut_operator
+class _coconut_attrgetter(object):
+    __slots__ = ("attrs",)
+    def __init__(self, *attrs):
+        self.attrs = attrs
+    def __reduce_ex__(self, _):
+        return self.__reduce__()
+    def __reduce__(self):
+        return (self.__class__, self.attrs)
+    @staticmethod
+    def _getattr(obj, attr):
+        for name in attr.split("."):
+            obj = _coconut.getattr(obj, name)
+        return obj
+    def __call__(self, obj):
+        if len(self.attrs) == 1:
+            return self._getattr(obj, self.attrs[0])
+        return _coconut.tuple(self._getattr(obj, attr) for attr in self.attrs)
+_coconut_operator.attrgetter = _coconut_attrgetter
+class _coconut_itemgetter(object):
+    __slots__ = ("items",)
+    def __init__(self, *items):
+        self.items = items
+    def __reduce_ex__(self, _):
+        return self.__reduce__()
+    def __reduce__(self):
+        return (self.__class__, self.items)
+    def __call__(self, obj):
+        if len(self.items) == 1:
+            return obj[self.items[0]]
+        return _coconut.tuple(obj[item] for item in self.items)
+_coconut_operator.itemgetter = _coconut_itemgetter
+class _coconut_methodcaller(object):
+    __slots__ = ("name", "args", "kwargs")
+    def __init__(self, name, *args, **kwargs):
+        self.name = name
+        self.args = args
+        self.kwargs = kwargs
+    def __reduce_ex__(self, _):
+        return self.__reduce__()
+    def __reduce__(self):
+        return (self.__class__, (self.name,) + self.args, {"kwargs": self.kwargs})
+    def __setstate__(self, setvars):
+        for k, v in setvars.items():
+            _coconut.setattr(self, k, v)
+    def __call__(self, obj):
+        return _coconut.getattr(obj, self.name)(*self.args, **self.kwargs)
+_coconut_operator.methodcaller = _coconut_methodcaller
+'''
 
-PY2_HEADER = PY27_HEADER + '''if _coconut_sys.version_info < (2, 7):
+_non_py37_extras = r'''def _coconut_default_breakpointhook(*args, **kwargs):
+    hookname = _coconut.os.getenv("PYTHONBREAKPOINT")
+    if hookname != "0":
+        if not hookname:
+            hookname = "pdb.set_trace"
+        modname, dot, funcname = hookname.rpartition(".")
+        if not dot:
+            modname = "builtins" if _coconut_sys.version_info >= (3,) else "__builtin__"
+        if _coconut_sys.version_info >= (2, 7):
+            import importlib
+            module = importlib.import_module(modname)
+        else:
+            import imp
+            module = imp.load_module(modname, *imp.find_module(modname))
+        hook = _coconut.getattr(module, funcname)
+        return hook(*args, **kwargs)
+if not hasattr(_coconut_sys, "__breakpointhook__"):
+    _coconut_sys.__breakpointhook__ = _coconut_default_breakpointhook
+def breakpoint(*args, **kwargs):
+    return _coconut.getattr(_coconut_sys, "breakpointhook", _coconut_default_breakpointhook)(*args, **kwargs)
+'''
+
+_finish_dict_def = '''
+    def __or__(self, other):
+        out = self.copy()
+        out.update(other)
+        return out
+    def __ror__(self, other):
+        out = self.__class__(other)
+        out.update(self)
+        return out
+    def __ior__(self, other):
+        self.update(other)
+        return self
+class _coconut_dict_meta(type):
+    def __instancecheck__(cls, inst):
+        return _coconut.isinstance(inst, _coconut_py_dict)
+    def __subclasscheck__(cls, subcls):
+        return _coconut.issubclass(subcls, _coconut_py_dict)
+dict = _coconut_dict_meta(py_str("dict"), _coconut_dict_base.__bases__, _coconut_dict_base.__dict__.copy())
+'''
+
+_below_py37_extras = '''from collections import OrderedDict as _coconut_OrderedDict
+class _coconut_dict_base(_coconut_OrderedDict):
+    __slots__ = ()
+    __doc__ = getattr(_coconut_OrderedDict, "__doc__", "<see help(py_dict)>")
+    __eq__ = _coconut_py_dict.__eq__
+    def __repr__(self):
+        return "{" + ", ".join("{k!r}: {v!r}".format(k=k, v=v) for k, v in self.items()) + "}"''' + _finish_dict_def
+
+_py37_py38_extras = '''class _coconut_dict_base(_coconut_py_dict):
+    __slots__ = ()
+    __doc__ = getattr(_coconut_py_dict, "__doc__", "<see help(py_dict)>")''' + _finish_dict_def
+
+_py26_extras = '''if _coconut_sys.version_info < (2, 7):
     import functools as _coconut_functools, copy_reg as _coconut_copy_reg
     def _coconut_new_partial(func, args, keywords):
         return _coconut_functools.partial(func, *(args if args is not None else ()), **(keywords if keywords is not None else {}))
@@ -250,9 +322,90 @@ PY2_HEADER = PY27_HEADER + '''if _coconut_sys.version_info < (2, 7):
     _coconut_copy_reg.pickle(_coconut_functools.partial, _coconut_reduce_partial)
 '''
 
-PYCHECK_HEADER = r'''if _coconut_sys.version_info < (3,):
-''' + _indent(PY2_HEADER) + '''else:
-''' + _indent(PY3_HEADER)
+_py3_before_py311_extras = '''try:
+    from exceptiongroup import ExceptionGroup, BaseExceptionGroup
+except ImportError:
+    class you_need_to_install_exceptiongroup(object):
+        __slots__ = ()
+    ExceptionGroup = BaseExceptionGroup = you_need_to_install_exceptiongroup()
+'''
+
+
+# whenever new versions are added here, header.py must be updated to use them
+ROOT_HEADER_VERSIONS = (
+    "universal",
+    "2",
+    "27",
+    "3",
+    "37",
+    "39",
+    "311",
+)
+
+
+def _get_root_header(version="universal"):
+    assert version in ROOT_HEADER_VERSIONS, version
+
+    if version == "universal":
+        return r'''if _coconut_sys.version_info < (3,):
+''' + _indent(_get_root_header("2")) + '''else:
+''' + _indent(_get_root_header("3"))
+
+    version_info = _get_target_info(version)
+    header = ""
+
+    if version.startswith("3"):
+        header += _base_py3_header
+    else:
+        assert version.startswith("2"), version
+        # if a new assignment is added below, a new builtins import should be added alongside it
+        header += _base_py2_header
+
+    if version_info >= (3, 7):
+        header += r'''py_breakpoint = breakpoint
+'''
+    elif version == "3":
+        header += r'''if _coconut_sys.version_info < (3, 7):
+''' + _indent(_non_py37_extras) + r'''else:
+    py_breakpoint = breakpoint
+'''
+    else:
+        assert version.startswith("2"), version
+        header += _non_py37_extras
+        if version == "2":
+            header += _py26_extras
+
+    if version == "3":
+        header += r'''if _coconut_sys.version_info < (3, 7):
+''' + _indent(_below_py37_extras) + r'''elif _coconut_sys.version_info < (3, 9):
+''' + _indent(_py37_py38_extras)
+    elif (3, 7) <= version_info < (3, 9):
+        header += r'''if _coconut_sys.version_info < (3, 9):
+''' + _indent(_py37_py38_extras)
+    elif version.startswith("2"):
+        header += _below_py37_extras + '''dict.keys = _coconut_OrderedDict.viewkeys
+dict.values = _coconut_OrderedDict.viewvalues
+dict.items = _coconut_OrderedDict.viewitems
+'''
+    else:
+        assert version_info >= (3, 9), version
+
+    if (3,) <= version_info < (3, 11):
+        header += r'''if _coconut_sys.version_info < (3, 11):
+''' + _indent(_py3_before_py311_extras)
+
+    return header
+
+
+# -----------------------------------------------------------------------------------------------------------------------
+# CONSTANTS:
+# -----------------------------------------------------------------------------------------------------------------------
+
+if DEVELOP:
+    VERSION += "-" + ("a" if ALPHA else "post") + "_dev" + str(int(DEVELOP))
+VERSION_STR = VERSION + (" [" + VERSION_NAME + "]" if VERSION_NAME else "")
+
+PY2 = _coconut_sys.version_info < (3,)
 
 # -----------------------------------------------------------------------------------------------------------------------
 # SETUP:
@@ -269,11 +422,4 @@ _coconut.pickle = pickle
 import os
 _coconut.os = os
 
-if PY26:
-    exec(PY2_HEADER)
-elif PY2:
-    exec(PY27_HEADER)
-elif PY37:
-    exec(PY37_HEADER)
-else:
-    exec(PY3_HEADER)
+exec(_get_root_header())
